"""Rules to manage shading objects."""
import abc
import logging

import HABApp

import habapp_rules.actors.config.shading
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member, too-many-instance-attributes, too-many-locals
class _ShadingBase(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Base class for shading objects."""

	states = [
		{"name": "WindAlarm"},
		{"name": "Manual"},
		{"name": "Hand", "timeout": 20 * 3600, "on_timeout": "_auto_hand_timeout"},
		{"name": "Auto", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "Open"},
			{"name": "DoorOpen", "initial": "Open", "children": [
				{"name": "Open"},
				{"name": "PostOpen", "timeout": 5 * 60, "on_timeout": "_timeout_post_door_open"}
			]},
			{"name": "NightClose"},
			{"name": "SleepingClose"},
			{"name": "SunProtection"},
		]}
	]

	trans = [
		# wind alarm
		{"trigger": "_wind_alarm_on", "source": ["Auto", "Hand", "Manual"], "dest": "WindAlarm"},
		{"trigger": "_wind_alarm_off", "source": "WindAlarm", "dest": "Manual", "conditions": "_manual_active"},
		{"trigger": "_wind_alarm_off", "source": "WindAlarm", "dest": "Auto", "unless": "_manual_active"},

		# manual
		{"trigger": "_manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
		{"trigger": "_manual_off", "source": "Manual", "dest": "Auto"},

		# hand
		{"trigger": "_hand_command", "source": ["Auto"], "dest": "Hand"},
		{"trigger": "_auto_hand_timeout", "source": "Hand", "dest": "Auto"},

		# sun
		{"trigger": "_sun_on", "source": "Auto_Open", "dest": "Auto_SunProtection"},
		{"trigger": "_sun_off", "source": "Auto_SunProtection", "dest": "Auto_Open"},

		# sleep
		{"trigger": "_sleep_started", "source": ["Auto_Open", "Auto_NightClose", "Auto_SunProtection"], "dest": "Auto_SleepingClose"},
		{"trigger": "_sleep_started", "source": "Hand", "dest": "Auto"},
		{"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_SunProtection", "conditions": "_sun_protection_active_and_configured"},
		{"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_NightClose", "conditions": ["_night_active_and_configured"]},
		{"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_Open", "unless": ["_night_active_and_configured", "_sun_protection_active_and_configured"]},

		# door
		{"trigger": "_door_open", "source": ["Auto_NightClose", "Auto_SunProtection", "Auto_SleepingClose", "Auto_Open"], "dest": "Auto_DoorOpen"},
		{"trigger": "_door_open", "source": "Auto_DoorOpen_PostOpen", "dest": "Auto_DoorOpen_Open"},
		{"trigger": "_door_closed", "source": "Auto_DoorOpen_Open", "dest": "Auto_DoorOpen_PostOpen"},
		{"trigger": "_timeout_post_door_open", "source": "Auto_DoorOpen_PostOpen", "dest": "Auto_Init"},

		# night close
		{"trigger": "_night_started", "source": ["Auto_Open", "Auto_SunProtection"], "dest": "Auto_NightClose", "conditions": "_night_active_and_configured"},
		{"trigger": "_night_stopped", "source": "Auto_NightClose", "dest": "Auto_SunProtection", "conditions": "_sun_protection_active_and_configured"},
		{"trigger": "_night_stopped", "source": "Auto_NightClose", "dest": "Auto_Open", "unless": ["_sun_protection_active_and_configured"]}

	]
	_item_shading_position: HABApp.openhab.items.rollershutter_item.RollershutterItem | HABApp.openhab.items.dimmer_item.DimmerItem
	_state_observer_pos: habapp_rules.actors.state_observer.StateObserverRollerShutter | habapp_rules.actors.state_observer.StateObserverDimmer

	# pylint: disable=too-many-arguments
	def __init__(self,
	             name_shading_position: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.shading.ShadingConfig,
	             shading_position_control_names: list[str] | None = None,
	             shading_position_group_names: list[str] | None = None,
	             name_wind_alarm: str | None = None,
	             name_sun_protection: str | None = None,
	             name_sleeping_state: str | None = None,
	             name_night: str | None = None,
	             name_door: str | None = None,
	             name_summer: str | None = None,
	             name_hand_manual_is_active_feedback: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None,
	             value_tolerance: int = 0) -> None:
		"""Init of _ShadingBase.

		:param name_shading_position:  name of OpenHAB shading item (RollershutterItem | DimmerItem)
		:param name_manual: name of OpenHAB switch item to disable all automatic functions
		:param config: configuration of the shading object
		:param shading_position_control_names: [optional] list of control items.
		:param shading_position_group_names: [optional]  list of group items where the item is a part of. Group item type must match with type of item_name
		:param name_wind_alarm: [optional] name of OpenHAB switch item which is 'ON' when wind-alarm is active
		:param name_sun_protection: [optional] name of OpenHAB switch item which is 'ON' when sun protection is active
		:param name_sleeping_state: [optional] name of OpenHAB sleeping state item (StringItem)
		:param name_night: [optional] name of OpenHAB switch item which is 'ON' when it is dark and 'OFF' when it is bright
		:param name_door: [optional] name of OpenHAB door item (ContactItem)
		:param name_summer: [optional] name of OpenHAB switch item which is 'ON' during summer and 'OFF' during winter
		:param name_hand_manual_is_active_feedback: [optional] name of OpenHAB switch item which is 'ON' if state is manual or hand
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: [optional] label of OpenHAB item for storing the current state (StringItem)
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if given config / items are not valid
		"""
		self._config = config
		self._value_tolerance = value_tolerance

		if not name_state:
			name_state = f"H_{name_shading_position}_state"
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, name_state, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_shading_position)

		# init items
		self._item_shading_position = HABApp.openhab.items.base_item.OpenhabItem.get_item(name_shading_position)
		self._item_manual = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_manual)
		self._item_wind_alarm = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_wind_alarm) if name_wind_alarm else None
		self._item_sun_protection = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_sun_protection) if name_sun_protection else None
		self._item_sleeping_state = HABApp.openhab.items.string_item.StringItem.get_item(name_sleeping_state) if name_sleeping_state else None
		self._item_night = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_night) if name_night else None
		self._item_door = HABApp.openhab.items.contact_item.ContactItem.get_item(name_door) if name_door else None
		self._item_summer = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_summer) if name_summer else None
		self._item_hand_manual_is_active_feedback = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_hand_manual_is_active_feedback) if name_hand_manual_is_active_feedback else None

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()
		self._check_config()
		self._apply_config()

		self._position_before: habapp_rules.actors.config.shading.ShadingPosition = habapp_rules.actors.config.shading.ShadingPosition(self._item_shading_position.value)

		if isinstance(self._item_shading_position, HABApp.openhab.items.rollershutter_item.RollershutterItem):
			self._state_observer_pos = habapp_rules.actors.state_observer.StateObserverRollerShutter(name_shading_position, self._cb_hand, shading_position_control_names, shading_position_group_names, value_tolerance)
		elif isinstance(self._item_shading_position, HABApp.openhab.items.dimmer_item.DimmerItem):
			self._state_observer_pos = habapp_rules.actors.state_observer.StateObserverDimmer(name_shading_position, self._cb_hand, self._cb_hand, self._cb_hand, shading_position_control_names, shading_position_group_names, value_tolerance)
		else:
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"shading position item must be a dimmer or roller-shutter item. Given type: {type(self._item_shading_position)}")

		# callbacks
		self._item_manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_wind_alarm is not None:
			self._item_wind_alarm.listen_event(self._cb_wind_alarm, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_sun_protection is not None:
			self._item_sun_protection.listen_event(self._cb_sun, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_sleeping_state is not None:
			self._item_sleeping_state.listen_event(self._cb_sleep_state, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_night is not None:
			self._item_night.listen_event(self._cb_night, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_door is not None:
			self._item_door.listen_event(self._cb_door, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._update_openhab_state()

	def _check_config(self):
		"""Check if config can be applied with the given items.

		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if given config / items are not valid
		"""
		if self._config.pos_night_close_summer is not None and self._item_summer is None:
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Night close position is set for summer, but item for summer / winter is missing!")

	def _apply_config(self) -> None:
		"""Apply config to state machine."""
		# set timeouts
		self.state_machine.states["Auto"].states["DoorOpen"].states["PostOpen"].timeout = self._config.door_post_time
		self.state_machine.states["Manual"].timeout = self._config.manual_timeout

	def _get_initial_state(self, default_value: str = "") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_wind_alarm is not None and self._item_wind_alarm.is_on():
			return "WindAlarm"
		if self._item_manual.is_on():
			return "Manual"
		if self._item_door is not None and self._item_door.is_open():  # self._item_door.is_open():
			return "Auto_DoorOpen_Open"
		if self._item_sleeping_state in (habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value):
			return "Auto_SleepingClose"
		if self._item_night is not None and self._item_night.is_on() and self._night_active_and_configured():
			return "Auto_NightClose"
		if self._sun_protection_active_and_configured():
			return "Auto_SunProtection"
		return "Auto_Open"

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item and other states.

		This method should be set to "after_state_change" of the state machine.
		"""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

			self._set_shading_state()

			if self._item_hand_manual_is_active_feedback is not None:
				self._item_hand_manual_is_active_feedback.oh_post_update("ON" if self.state in {"Manual", "Hand"} else "OFF")

			self._previous_state = self.state

	@abc.abstractmethod
	def _set_shading_state(self) -> None:
		"""Set shading state"""

	def _get_target_position(self) -> habapp_rules.actors.config.shading.ShadingPosition | None:
		"""Get target position for shading object.

		:return: target shading position
		"""

		if self.state in {"Hand", "Manual"}:
			if self._previous_state == "WindAlarm":
				return self._position_before
			return None

		if self.state == "WindAlarm":
			return self._config.pos_wind_alarm

		if self.state == "Auto_Open":
			return self._config.pos_auto_open

		if self.state == "Auto_SunProtection":
			return self._config.pos_sun_protection

		if self.state == "Auto_SleepingClose":
			return self._config.pos_sleeping

		if self.state == "Auto_NightClose":
			if self._item_summer.is_on():
				return self._config.pos_night_close_summer
			return self._config.pos_night_close_winter

		if self.state == "Auto_DoorOpen_Open":
			return self._config.pos_door_open

		return None

	def on_enter_Auto_Init(self) -> None:  # pylint: disable=invalid-name
		"""Is called on entering of init state"""
		self._set_initial_state()

	def on_exit_Manual(self) -> None:  # pylint: disable=invalid-name
		"""Is called if state Manual is left."""
		self._set_position_before()

	def on_exit_Hand(self) -> None:  # pylint: disable=invalid-name
		"""Is called if state Hand is left."""
		self._set_position_before()

	def _set_position_before(self) -> None:
		"""Set / save position before manual state is entered. This is used to restore the previous position"""
		self._position_before = habapp_rules.actors.config.shading.ShadingPosition(self._item_shading_position.value)

	def _manual_active(self) -> bool:
		"""Check if manual is active.

		:return: True if night is active
		"""
		return self._item_manual.is_on()

	def _sun_protection_active_and_configured(self) -> bool:
		"""Check if sun protection is active.

		:return: True if night is active
		"""
		return self._item_sun_protection is not None and self._item_sun_protection.is_on() and self._config.pos_sun_protection is not None

	def _night_active_and_configured(self) -> bool:
		"""Check if night is active and configured.

		:return: True if night is active
		"""
		night_config = self._config.pos_night_close_summer if self._item_summer is not None and self._item_summer.is_on() else self._config.pos_night_close_winter
		return self._item_night is not None and self._item_night.is_on() and night_config is not None

	def _cb_hand(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if a external control was detected.

		:param event: original trigger event
		"""
		self._hand_command()

	def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if manual mode changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._manual_on()
		else:
			self._manual_off()

	def _cb_wind_alarm(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if wind alarm changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._wind_alarm_on()
		else:
			self._wind_alarm_off()

	def _cb_sun(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if sun state changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._sun_on()
		else:
			self._sun_off()

	def _cb_sleep_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if sleeping state changed.

		:param event: original trigger event
		"""
		if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
			self._sleep_started()
		elif event.value == habapp_rules.system.SleepState.POST_SLEEPING.value:
			self._sleep_stopped()

	def _cb_night(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if night / dark state changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._night_started()
		else:
			self._night_stopped()

	def _cb_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if door state changed.

		:param event: original trigger event
		"""
		if event.value == "OPEN":
			self._door_open()
		else:
			self._door_closed()


class Shutter(_ShadingBase):
	"""Rules class to manage a normal shutters (or curtains).

			# KNX-things:
			Thing device KNX_Shading "KNX OpenHAB dimmer observer"{
				Type dimmer                 : shading_position          "Shading position"          [ position="5.001:4/1/12+<4/1/15" ]
	            Type dimmer-control         : shading_position_ctr      "Shading position ctr"      [ position="5.001:4/1/12+<4/1/15" ]
				Type dimmer-control         : shading_group_all_ctr     "Shading all ctr"           [ position="5.001:4/1/112+<4/1/115""]
				Type switch-control         : shading_hand_manual_ctr   "Shading hand / manual"     [ ga="4/1/20" ]
			}

			# Items:
			Rollershutter    shading_position       "Shading position [%s %%]"          <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position"}
			Rollershutter    shading_position_ctr   "Shading position ctr [%s %%]"      <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position_ctr"}
			Switch           shading_manual         "Shading manual"
			Rollershutter    shading_all_ctr        "Shading all ctr [%s %%]"           <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_group_all_ctr"}
			Switch           shading_hand_manual    "Shading in Hand / Manual state"                        {channel="knx:device:bridge:KNX_Shading:shading_hand_manual_ctr"}


			# Rule init:
			habapp_rules.actors.shading.Shutter(
				"shading_position",
				"shading_manual",
				habapp_rules.actors.config.shading.CONFIG_DEFAULT,
				["shading_position_ctr", "shading_all_ctr"],
				[],
				"I99_99_WindAlarm",
				"I99_99_SunProtection",
				"I99_99_Sleeping_State",
				"I99_99_Night",
				"I99_99_Door",
				"I99_99_Summer"
				"shading_hand_manual"
			)
			"""

	# pylint: disable=too-many-arguments,too-many-locals
	def __init__(self,
	             name_shading_position: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.shading.ShadingConfig,
	             shading_position_control_names: list[str] | None = None,
	             shading_position_group_names: list[str] | None = None,
	             name_wind_alarm: str | None = None,
	             name_sun_protection: str | None = None,
	             name_sleeping_state: str | None = None,
	             name_night: str | None = None,
	             name_door: str | None = None,
	             name_summer: str | None = None,
	             name_hand_manual_is_active_feedback: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None,
	             value_tolerance: int = 0) -> None:
		"""Init of Raffstore object.

		:param name_shading_position: name of OpenHAB shading item (RollershutterItem)
		:param name_manual: name of OpenHAB switch item to disable all automatic functions
		:param config: configuration of the shading object
		:param shading_position_control_names: [optional] list of control items.
		:param shading_position_group_names: [optional]  list of group items where the item is a part of. Group item type must match with type of item_name
		:param name_wind_alarm: [optional] name of OpenHAB switch item which is 'ON' when wind-alarm is active
		:param name_sun_protection: [optional] name of OpenHAB switch item which is 'ON' when sun protection is active
		:param name_sleeping_state: [optional] name of OpenHAB sleeping state item (StringItem)
		:param name_night: [optional] name of OpenHAB switch item which is 'ON' when it is dark and 'OFF' when it is bright
		:param name_door: [optional] name of OpenHAB door item (ContactItem)
		:param name_summer: [optional] name of OpenHAB switch item which is 'ON' during summer and 'OFF' during winter
		:param name_hand_manual_is_active_feedback: [optional] name of OpenHAB switch item which is 'ON' if state is manual or hand
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: [optional] label of OpenHAB item for storing the current state (StringItem)
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		"""
		_ShadingBase.__init__(
			self,
			name_shading_position,
			name_manual,
			config,
			shading_position_control_names,
			shading_position_group_names,
			name_wind_alarm, name_sun_protection,
			name_sleeping_state,
			name_night, name_door,
			name_summer,
			name_hand_manual_is_active_feedback,
			name_state,
			state_label,
			value_tolerance
		)

		self._instance_logger.debug(self.get_initial_log_message())

	def _set_shading_state(self) -> None:
		"""Set shading state"""
		if self._previous_state is None:
			# don't change value if called during init (_previous_state == None)
			return

		target_position = self._get_target_position()

		if target_position is None:
			return

		if (position := target_position.position) is not None:
			self._state_observer_pos.send_command(position)
			self._instance_logger.debug(f"set position {target_position.position}")


# pylint: disable=too-many-arguments
class Raffstore(_ShadingBase):
	"""Rules class to manage a raffstore.

		# KNX-things:
		Thing device KNX_Shading "KNX OpenHAB dimmer observer"{
			Type rollershutter          : shading_position          "Shading position"          [ upDown="4/1/10", stopMove="4/1/11", position="5.001:4/1/12+<4/1/15" ]
            Type rollershutter-control  : shading_position_ctr      "Shading position ctr"      [ upDown="4/1/10", stopMove="4/1/11" ]
            Type dimmer                 : shading_slat              "Shading slat"              [ position="5.001:4/1/13+<4/1/16" ]
			Type rollershutter-control  : shading_group_all_ctr     "Shading all ctr"           [ upDown="4/1/110", stopMove="4/1/111"]
			Type switch-control         : shading_hand_manual_ctr   "Shading hand / manual"     [ ga="4/1/20" ]
		}

		# Items:
		Rollershutter    shading_position       "Shading position [%s %%]"          <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position"}
		Rollershutter    shading_position_ctr   "Shading position ctr [%s %%]"      <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position_ctr"}
		Dimmer           shading_slat           "Shading slat [%s %%]"              <slat>              {channel="knx:device:bridge:KNX_Shading:shading_slat"}
		Switch           shading_manual         "Shading manual"
		Rollershutter    shading_all_ctr        "Shading all ctr [%s %%]"           <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_group_all_ctr"}
		Switch           shading_hand_manual    "Shading in Hand / Manual state"                        {channel="knx:device:bridge:KNX_Shading:shading_hand_manual_ctr"}


		# Rule init:
		habapp_rules.actors.shading.Raffstore(
			"shading_position",
			"shading_slat",
			"shading_manual",
			habapp_rules.actors.config.shading.CONFIG_DEFAULT,
			["shading_position_ctr", "shading_all_ctr"],
			[],
			"I99_99_WindAlarm",
			"I99_99_SunProtection",
			"I99_99_SunProtection_Slat",
			"I99_99_Sleeping_State",
			"I99_99_Night",
			"I99_99_Door",
			"I99_99_Summer"
			"shading_hand_manual"
		)
		"""

	# pylint: disable=too-many-locals
	def __init__(self,
	             name_shading_position: str,
	             name_slat: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.shading.ShadingConfig,
	             shading_position_control_names: list[str] | None = None,
	             shading_position_group_names: list[str] | None = None,
	             name_wind_alarm: str | None = None,
	             name_sun_protection: str | None = None,
	             name_sun_protection_slat: str | None = None,
	             name_sleeping_state: str | None = None,
	             name_night: str | None = None,
	             name_door: str | None = None,
	             name_summer: str | None = None,
	             name_hand_manual_is_active_feedback: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None,
	             value_tolerance: int = 0) -> None:
		"""Init of Raffstore object.

		:param name_shading_position: name of OpenHAB shading item (RollershutterItem)
		:param name_slat: name of OpenHAB slat item (DimmerItem)
		:param name_manual: name of OpenHAB switch item to disable all automatic functions
		:param config: configuration of the shading object
		:param shading_position_control_names: [optional] list of control items.
		:param shading_position_group_names: [optional]  list of group items where the item is a part of. Group item type must match with type of item_name
		:param name_wind_alarm: [optional] name of OpenHAB switch item which is 'ON' when wind-alarm is active
		:param name_sun_protection: [optional] name of OpenHAB switch item which is 'ON' when sun protection is active
		:param name_sun_protection_slat:
		:param name_sleeping_state: [optional] name of OpenHAB sleeping state item (StringItem)
		:param name_night: [optional] name of OpenHAB switch item which is 'ON' when it is dark and 'OFF' when it is bright
		:param name_door: [optional] name of OpenHAB door item (ContactItem)
		:param name_summer: [optional] name of OpenHAB switch item which is 'ON' during summer and 'OFF' during winter
		:param name_hand_manual_is_active_feedback: [optional] name of OpenHAB switch item which is 'ON' if state is manual or hand
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: [optional] label of OpenHAB item for storing the current state (StringItem)
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		"""
		_ShadingBase.__init__(
			self,
			name_shading_position,
			name_manual,
			config,
			shading_position_control_names,
			shading_position_group_names,
			name_wind_alarm, name_sun_protection,
			name_sleeping_state,
			name_night, name_door,
			name_summer,
			name_hand_manual_is_active_feedback,
			name_state,
			state_label,
			value_tolerance
		)

		self._state_observer_slat = habapp_rules.actors.state_observer.StateObserverSlat(name_slat, self._cb_hand, value_tolerance)

		# init items
		self._item_slat = HABApp.openhab.items.dimmer_item.DimmerItem.get_item(name_slat)
		self._item_sun_protection_slat = HABApp.openhab.items.dimmer_item.DimmerItem.get_item(name_sun_protection_slat) if name_sun_protection_slat else None
		self.__verify_items()

		# callbacks
		if self._item_sun_protection_slat is not None:
			self._item_sun_protection_slat.listen_event(self._cb_slat_target, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(self.get_initial_log_message())

	def __verify_items(self) -> None:
		"""Check if given items are valid

		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if given items are not valid
		"""
		# check type of rollershutter item
		if not isinstance(self._item_shading_position, HABApp.openhab.items.rollershutter_item.RollershutterItem):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"The shading position item must be of type RollershutterItem. Given: {type(self._item_shading_position)}")

		# check if the correct items are given for sun protection mode
		if (self._item_sun_protection is None) != (self._item_sun_protection_slat is None):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Ether sun_protection item AND sun_protection_slat item must be given or None of them.")

	def _get_target_position(self) -> habapp_rules.actors.config.shading.ShadingPosition | None:
		"""Get target position for shading object(s).

		:return: target shading position
		"""
		target_position = super()._get_target_position()

		if self.state == "Auto_SunProtection" and target_position is not None:
			target_position.slat = self._item_sun_protection_slat.value

		return target_position

	def _set_shading_state(self) -> None:
		"""Set shading state"""
		if self._previous_state is None:
			# don't change value if called during init (_previous_state == None)
			return

		target_position = self._get_target_position()

		if target_position is None:
			return

		if (position := target_position.position) is not None:
			self._state_observer_pos.send_command(position)

		if (slat := target_position.slat) is not None:
			self._state_observer_slat.send_command(slat)

		if any(pos is not None for pos in (position, slat)):
			self._instance_logger.debug(f"set position {target_position}")

	def _set_position_before(self) -> None:
		"""Set / save position before manual state is entered. This is used to restore the previous position"""
		self._position_before = habapp_rules.actors.config.shading.ShadingPosition(self._item_shading_position.value, self._item_slat.value)

	def _cb_slat_target(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if the target slat value changed.

		:param event: original trigger event
		"""
		if self.state == "Auto_SunProtection":
			self._state_observer_slat.send_command(event.value)


class ResetAllManualHand(HABApp.Rule):
	"""Clear the state hand / manual state of all shading

	# Items:
	Switch           clear_hand_manual         "Clear Hand / Manual state of all shading objects"

	# Rule init:
	habapp_rules.actors.shading.ResetAllManualHand("clear_hand_manual")
	"""

	def __init__(self, name_reset_manual_hand: str) -> None:
		"""Init of reset class.

		:param name_reset_manual_hand: name of OpenHAB reset item (SwitchItem)
		"""
		HABApp.Rule.__init__(self)
		self._item_reset = HABApp.openhab.items.SwitchItem.get_item(name_reset_manual_hand)
		self._item_reset.listen_event(self._cb_reset_all, HABApp.openhab.events.ItemStateUpdatedEventFilter())

	def __get_shading_objects(self) -> list[type(_ShadingBase)]:
		"""Get all shading objects.

		:return: list of shading objects
		"""
		return [rule for rule in self.get_rule(None) if issubclass(rule.__class__, _ShadingBase)]

	def _cb_reset_all(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback which is called if reset is requested

		:param event: trigger event
		"""
		if event.value == "OFF":
			return

		for shading_object in self.__get_shading_objects():
			state = shading_object.state
			manual_item: HABApp.openhab.items.SwitchItem = shading_object._item_manual  # pylint: disable=protected-access

			if state == "Manual":
				manual_item.oh_send_command("OFF")

			elif state == "Hand":
				manual_item.oh_send_command("ON")
				manual_item.oh_send_command("OFF")

		self._item_reset.oh_send_command("OFF")


class SlatValueSun(HABApp.Rule):
	"""Rules class to get slat value depending on sun elevation.

		# Items:
		Number    elevation             "Sun elevation [%s]"    <sun>     {channel="astro...}
		Number    sun_protection_slat   "Slat value [%s %%]"    <slat>

		# Rule init:
		habapp_rules.actors.shading.SlatValueSun(
			"elevation",
			"sun_protection_slat",
			habapp_rules.actors.config.shading.CONFIG_DEFAULT_ELEVATION_SLAT_WINTER,
			"I99_99_Summer",
			habapp_rules.actors.config.shading.CONFIG_DEFAULT_ELEVATION_SLAT_SUMMER,
		)
		"""

	_item_slat_value: HABApp.openhab.items.dimmer_item.DimmerItem | HABApp.openhab.items.number_item.NumberItem

	def __init__(self,
	             name_sun_elevation: str,
	             name_slat_value: str,
	             elevation_slat_characteristic: list[(float | int, float | int)],
	             name_summer: str | None = None,
	             elevation_slat_characteristic_summer: list[(float | int, float | int)] | None = None):
		"""Init SlatValueSun

		:param name_sun_elevation: name of OpenHAB sun elevation item (NumberItem)
		:param name_slat_value: name of OpenHAB target slat item (NumberItem | DimmerItem)
		:param elevation_slat_characteristic: List of tuple-mappings from elevation to slat value
		:param name_summer: name of OpenHAB summer item (SwitchItem)
		:param elevation_slat_characteristic_summer: List of tuple-mappings from elevation to slat value, which is used if summer is active
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not valid
		"""
		if (name_summer is None) != (elevation_slat_characteristic_summer is None):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Ether both 'name_summer' and 'elevation_slat_characteristic_summer' must be None or a value.")

		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_slat_value)

		# init items
		self._item_sun_elevation = HABApp.openhab.items.number_item.NumberItem.get_item(name_sun_elevation)
		self._item_slat_value = HABApp.openhab.items.OpenhabItem.get_item(name_slat_value)
		self._item_summer = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_summer) if name_summer else None

		if not isinstance(self._item_slat_value, (HABApp.openhab.items.dimmer_item.DimmerItem, HABApp.openhab.items.number_item.NumberItem)):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Item slat_value must be of type DimmerItem or NumberItem!")

		# slat characteristics
		self.__slat_characteristic_default = self.__check_and_sort_characteristic(elevation_slat_characteristic)
		self.__slat_characteristic_summer = self.__check_and_sort_characteristic(elevation_slat_characteristic_summer) if elevation_slat_characteristic_summer else None
		self._slat_characteristic_active = self.__slat_characteristic_summer if self._item_summer is not None and self._item_summer.is_on() else self.__slat_characteristic_default

		# callbacks
		self._item_sun_elevation.listen_event(self._cb_elevation, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_summer is not None:
			self._item_summer.listen_event(self._cb_summer_winter, HABApp.openhab.events.ItemStateChangedEventFilter())
		self.run.soon(self.__send_slat_value)

		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with name '{self.rule_name}' was successful.")

	@staticmethod
	def __check_and_sort_characteristic(characteristic: list[(float | int, float | int)]) -> list[(float, float)]:
		"""Check and sort characteristic

		:param characteristic: characteristic for slats. First value must be elevation, second the slat value
		:return: sorted characteristic
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if characteristic is not valid
		"""

		if not isinstance(characteristic, list) or not all(isinstance(itm, tuple) and len(itm) == 2 for itm in characteristic):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Characteristic must be a list of value pairs as tuple")

		if len(set(value_pair[0] for value_pair in characteristic)) != len(characteristic):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Elevation values must be unique!")

		return sorted(characteristic)

	def __get_slat_value(self, elevation: float) -> float:
		"""Get slat value for given elevation.

		:param elevation: elevation of the sun
		:return: slat value
		"""
		if elevation >= self._slat_characteristic_active[-1][0]:
			return self._slat_characteristic_active[-1][1]
		if elevation < self._slat_characteristic_active[0][0]:
			return self._slat_characteristic_active[0][1]

		# no cover because of loop does not finish, but is handled with the two if statements above
		return next(config for idx, config in enumerate(self._slat_characteristic_active) if config[0] <= elevation < self._slat_characteristic_active[idx + 1][0])[1]  # pragma: no cover

	def __send_slat_value(self) -> None:
		"""Send slat value to OpenHAB item."""
		slat_value = self.__get_slat_value(self._item_sun_elevation.value)

		if self._item_slat_value.value != slat_value:
			self._item_slat_value.oh_send_command(slat_value)

	def _cb_elevation(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is called if sun elevation changed

		:param event: elevation event
		"""
		self.__send_slat_value()

	def _cb_summer_winter(self, event: HABApp.openhab.events.ItemStateChangedEvent):
		"""Callback which is called if summer / winter changed

		:param event: summer / winter event
		"""
		self._slat_characteristic_active = self.__slat_characteristic_summer if event.value == "ON" else self.__slat_characteristic_default
		self.__send_slat_value()



"""Common helper functions for all rules."""
import logging
import time

import HABApp
import HABApp.openhab.connection.handler.func_sync
import HABApp.openhab.items

import habapp_rules.core.exceptions

LOGGER = logging.getLogger(__name__)


def create_additional_item(name: str, item_type: str, label: str | None = None) -> HABApp.openhab.items.OpenhabItem:
	"""Create additional item if it does not already exist

	:param name: Name of item
	:param item_type: Type of item (e.g. String)
	:param label: Label of the item
	:return: returns the created item
	:raises habapp_rules.core.exceptions.HabAppRulesException: if item could not be created
	"""
	name = f"H_{name.removeprefix('H_')}"

	if not HABApp.openhab.interface_sync.item_exists(name):
		if not label:
			label = f"{name.removeprefix('H_').replace('_', ' ')}"
		if item_type == "String" and not label.endswith("[%s]"):
			label = f"{label} [%s]"
		if not HABApp.openhab.interface_sync.create_item(item_type=item_type, name=name, label=label):
			raise habapp_rules.core.exceptions.HabAppRulesException(f"Could not create item '{name}'")
		time.sleep(0.05)
	return HABApp.openhab.items.OpenhabItem.get_item(name)


def send_if_different(item: str | HABApp.openhab.items.OpenhabItem, value: str | float | int) -> None:
	"""Send command if the target value is different to the current value.

	:param item: name of OpenHab item
	:param value: value to write to OpenHAB item
	"""
	if isinstance(item, str):
		item = HABApp.openhab.items.OpenhabItem.get_item(item)

	if item.value != value:
		item.oh_send_command(value)


def filter_updated_items(input_items: list[HABApp.openhab.items.OpenhabItem], filter_time: int | None = None) -> list[HABApp.openhab.items.OpenhabItem]:
	"""Get input items depending on their last update time and _ignore_old_values_time

	:param input_items: all items which should be checked for last update time
	:param filter_time: threshold for last update time
	:return: full list if _ignore_old_values is not set, otherwise all items where updated in time.
	"""
	if filter_time is None:
		return input_items

	current_time = time.time()
	filtered_items = [item for item in input_items if current_time - item.last_update.timestamp() <= filter_time]

	if len(input_items) != len(filtered_items):
		ignored_item_names = [item.name for item in input_items if current_time - item.last_update.timestamp() > filter_time]
		LOGGER.warning(f"The following items where not updated during the last {filter_time}s and will be ignored: {ignored_item_names}")

	return filtered_items



"""Config for HCL color"""
import dataclasses


@dataclasses.dataclass
class LightHclConfig:
	"""Config for HCL color"""
	color_config: list[tuple[float, float]]
	hand_timeout: int = 5 * 3600  # 5 hours
	sleep_color: float | None = None
	post_sleep_timeout: int | None = None
	focus_color: float | None = None
	shift_weekend_holiday: bool = False  # only relevant for HclTime

	def __post_init__(self) -> None:
		"""Trigger checks after init."""
		self._sorted_color_config()

	def _sorted_color_config(self) -> None:
		"""Sort color config"""
		self.color_config = sorted(self.color_config, key=lambda x: x[0])


EXAMPLE_CONFIG_ELEVATION = LightHclConfig([
	(-15, 3900),
	(0, 4500),
	(5, 5500),
	(15, 6500)
])

EXAMPLE_CONFIG_TIME = LightHclConfig([
	(0, 2200),
	(4, 2200),
	(5, 3200),
	(6, 3940),
	(8, 5000),
	(12, 7000),
	(19, 7000),
	(21, 5450),
	(22, 4000),
	(23, 2600),
])



"""Exceptions for HabAppRules."""


class HabAppRulesException(Exception):
	"""Exception which is raised by this package."""


class HabAppRulesConfigurationException(HabAppRulesException):
	"""Exception which is raised if wrong configuration is given"""



"""Rules for bridging KNX controller to MQTT items."""
import logging

import HABApp

import habapp_rules.core.exceptions
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class KnxMqttDimmerBridge(HABApp.Rule):
	"""Create a bridge to control a MQTT dimmer from a KNX controller (e.g. wall switch).

	To use this the items must be configured according the following example:
	- mqtt_dimmer: autoupdate should be false, thing: according to OpenHAB documentation
	- knx_switch_ctr: autoupdate must be activated, thing:  [ ga="1/1/124+1/1/120" ] for ga: at first always use the RM-GA, second is the control-GA
	- knx_dimmer_ctr: autoupdate must be activated, thing:  [ position="1/1/125+1/1/123", increaseDecrease="1/1/122" ] for position: at first always use the RM-GA, second is the control-GA

	info: OpenHAB does not support start/stop dimming. Thus, this implementation will set fixed values if INCREASE/DECREASE was received from KNX
	"""

	def __init__(self, mqtt_dimmer: str, knx_switch_ctr: str | None = None, knx_dimmer_ctr: str | None = None, increase_value: int = 60, decrease_value: int = 30) -> None:
		"""Create object of KNX to MQTT bridge

		:param mqtt_dimmer: name of MQTT item
		:param knx_switch_ctr: name of KNX switch-control item
		:param knx_dimmer_ctr: name of KNX dimmer-control item
		:param increase_value: value which is set when INCREASE was received.
		:param decrease_value: value which is set when DECREASE was received.
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: If config is not valid
		"""
		if knx_switch_ctr is None and knx_dimmer_ctr is None:
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"At least one KNX item must be given! knx_switch_ctr = {knx_switch_ctr} | knx_dimmer_ctr = {knx_dimmer_ctr}")

		knx_name = knx_switch_ctr if knx_switch_ctr else knx_dimmer_ctr

		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{knx_name}__{mqtt_dimmer}")

		self.__increase_value = increase_value
		self.__decrease_value = decrease_value

		self._mqtt_item = HABApp.openhab.items.DimmerItem.get_item(mqtt_dimmer)
		self._knx_dimmer_item = HABApp.openhab.items.DimmerItem.get_item(knx_dimmer_ctr) if knx_dimmer_ctr else None
		self._knx_switch_item = HABApp.openhab.items.SwitchItem.get_item(knx_switch_ctr) if knx_switch_ctr else None

		self._mqtt_item.listen_event(self._cb_mqtt_event, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._knx_dimmer_item is not None:
			self._knx_dimmer_item.listen_event(self._cb_knx_event, HABApp.openhab.events.ItemCommandEventFilter())
		if self._knx_switch_item is not None:
			self._knx_switch_item.listen_event(self._cb_knx_event, HABApp.openhab.events.ItemCommandEventFilter())
		self._instance_logger.debug("successful!")

	def _cb_knx_event(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is called if a KNX command received.

		:param event: HABApp event
		"""
		if isinstance(event.value, (int, float)):
			self._mqtt_item.oh_send_command(event.value)
		elif event.value in {"ON", "OFF"}:
			self._mqtt_item.oh_send_command(event.value)
		elif event.value == "INCREASE":
			target_value = self.__increase_value if self._mqtt_item.value < self.__increase_value else 100
			self._mqtt_item.oh_send_command(target_value)
		elif event.value == "DECREASE":
			target_value = self.__decrease_value if self._mqtt_item.value > self.__decrease_value else 0
			self._mqtt_item.oh_send_command(target_value)
		else:
			self._instance_logger.error(f"command '{event.value}' ist not supported!")

	def _cb_mqtt_event(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if a MQTT state change event happens.

		:param event: HABApp event
		"""
		if self._knx_dimmer_item is not None:
			self._knx_dimmer_item.oh_post_update(event.value)

		if self._knx_switch_item is not None:
			self._knx_switch_item.oh_post_update("ON" if event.value > 0 else "OFF")



"""Rules to for garden watering."""
import datetime
import logging

import HABApp

import habapp_rules.core.exceptions
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class Irrigation(HABApp.Rule):
	"""Rule for easy irrigation control.

	# Items:
    Switch     I999_valve                   "Valve state"           {channel="some_channel_config"}
	Switch     I999_irrigation_active       "Irrigation active"
	Number     I999_irrigation_hour         "Start hour [%d]"
	Number     I999_irrigation_minute       "Start minute[%d]"
	Number     I999_irrigation_duration     "Duration [%d]"

	# Rule init:
	habapp_rules.actors.irrigation.Irrigation("I999_valve", "I999_irrigation_active", "I999_irrigation_hour", "I999_irrigation_minute", "I999_irrigation_duration")
	"""

	def __init__(self, name_valve: str, name_active: str, name_hour: str, name_minute: str, name_duration: str, name_repetitions: str | None = None, name_brake: str | None = None):
		"""Init of irrigation object.

		:param name_valve: name of OpenHAB valve item (SwitchItem)
		:param name_active: name of OpenHAB irrigation activation item (SwitchItem)
		:param name_hour: name of OpenHAB item for defining the start hour (NumberItem)
		:param name_minute: name of OpenHAB item for defining the start minute (NumberItem)
		:param name_duration: name of OpenHAB item for defining the duration (NumberItem)
		:param name_repetitions: [optional] name of OpenHAB item for defining the number of repetitions (NumberItem)
		:param name_brake: [optional] name of OpenHAB item for defining brake time in minutes between the repetitions (NumberItem)
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not correct
		"""

		if bool(name_repetitions) != bool(name_brake):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("If repeats item is given, also the brake item must be given!")

		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_valve)

		self._item_valve = HABApp.openhab.items.SwitchItem.get_item(name_valve)
		self._item_active = HABApp.openhab.items.SwitchItem.get_item(name_active)
		self._item_hour = HABApp.openhab.items.NumberItem.get_item(name_hour)
		self._item_minute = HABApp.openhab.items.NumberItem.get_item(name_minute)
		self._item_duration = HABApp.openhab.items.NumberItem.get_item(name_duration)
		self._item_repetitions = HABApp.openhab.items.NumberItem.get_item(name_repetitions) if name_repetitions else None
		self._item_brake = HABApp.openhab.items.NumberItem.get_item(name_brake) if name_brake else None

		self.run.soon(self._cb_set_valve_state)
		self.run.every_minute(self._cb_set_valve_state)

		self._item_active.listen_event(self._cb_set_valve_state, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._item_minute.listen_event(self._cb_set_valve_state, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._item_hour.listen_event(self._cb_set_valve_state, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_repetitions is not None and self._item_brake is not None:
			self._item_repetitions.listen_event(self._cb_set_valve_state, HABApp.openhab.events.ItemStateChangedEventFilter())  # pylint: disable=no-member
			self._item_brake.listen_event(self._cb_set_valve_state, HABApp.openhab.events.ItemStateChangedEventFilter())  # pylint: disable=no-member

		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with name '{self.rule_name}' was successful.")

	def _get_target_valve_state(self) -> bool:
		"""Get target valve state, depending on the OpenHAB item states

		:return: True if valve should be on, otherwise False
		:raises habapp_rules.core.exceptions.HabAppRulesException: if value for hour / minute / duration is not valid
		"""
		if not self._item_active.is_on():
			return False

		if any(item.value is None for item in (self._item_hour, self._item_minute, self._item_duration)):
			self._instance_logger.warning(
				f"OpenHAB item values are not valid for hour / minute / duration. Will return False. See current values: hour={self._item_hour.value} | minute={self._item_minute.value} | duration={self._item_duration.value}")
			return False

		repetitions = self._item_repetitions.value if self._item_repetitions else 0
		brake = int(self._item_brake.value) if self._item_brake else 0

		now = datetime.datetime.now()
		hour = int(self._item_hour.value)
		minute = int(self._item_minute.value)
		duration = int(self._item_duration.value)

		for idx in range(repetitions + 1):
			start_time = datetime.datetime.combine(date=now, time=datetime.time(hour, minute)) + datetime.timedelta(minutes=idx * (duration + brake))
			end_time = start_time + datetime.timedelta(minutes=duration)
			if self._is_in_time_range(start_time.time(), end_time.time(), now.time()):
				return True
		return False

	@staticmethod
	def _is_in_time_range(start_time: datetime.time, end_time: datetime.time, time_to_check: datetime.time) -> bool:
		"""Check if a time is in a given range.

		:param start_time: start time of the time range
		:param end_time: end time of the time range
		:param time_to_check: time, which should be checked
		:return: True if the time, which should be checked is between start and stop time
		"""
		if end_time < start_time:
			return start_time <= time_to_check or end_time > time_to_check
		return start_time <= time_to_check < end_time

	def _cb_set_valve_state(self, _: HABApp.openhab.events.ItemStateChangedEvent | None = None) -> None:
		"""Callback to set the valve state, triggered by cyclic call or item event"""
		try:
			target_value = self._get_target_valve_state()
		except habapp_rules.core.exceptions.HabAppRulesException as exc:
			self._instance_logger.warning(f"Could not get target valve state, set it to false. Error: {exc}")
			target_value = False

		if self._item_valve.is_on() != target_value:
			self._instance_logger.info(f"Valve state changed to {target_value}")
			self._item_valve.oh_send_command("ON" if target_value else "OFF")



"""Ventilation rules."""
import abc
import copy
import datetime
import logging

import HABApp

import habapp_rules.actors.config.ventilation
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class _VentilationBase(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Class for ventilation objects."""

	states = [
		{"name": "Manual"},
		{"name": "Auto", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "Normal"},
			{"name": "PowerHand", "timeout": 3600, "on_timeout": "_hand_off"},
			{"name": "PowerExternal"},
			{"name": "LongAbsence", "initial": "Off", "children": [
				{"name": "On", "timeout": 3600, "on_timeout": "_long_absence_power_off"},
				{"name": "Off"}
			]},
		]}
	]

	trans = [
		# manual
		{"trigger": "_manual_on", "source": ["Auto"], "dest": "Manual"},
		{"trigger": "_manual_off", "source": "Manual", "dest": "Auto"},

		# PowerHand
		{"trigger": "_hand_on", "source": ["Auto_Normal", "Auto_PowerExternal", "Auto_LongAbsence"], "dest": "Auto_PowerHand"},
		{"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_PowerExternal", "conditions": "_external_active_and_configured"},
		{"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_Normal", "unless": "_external_active_and_configured"},

		# PowerExternal
		{"trigger": "_external_on", "source": "Auto_Normal", "dest": "Auto_PowerExternal"},
		{"trigger": "_external_off", "source": "Auto_PowerExternal", "dest": "Auto_Normal"},

		# long absence
		{"trigger": "_long_absence_on", "source": ["Auto_Normal", "Auto_PowerExternal"], "dest": "Auto_LongAbsence"},
		{"trigger": "_long_absence_power_on", "source": "Auto_LongAbsence_Off", "dest": "Auto_LongAbsence_On"},
		{"trigger": "_long_absence_power_off", "source": "Auto_LongAbsence_On", "dest": "Auto_LongAbsence_Off"},
		{"trigger": "_long_absence_off", "source": "Auto_LongAbsence", "dest": "Auto_Normal"},
	]

	# pylint: disable=too-many-arguments
	def __init__(self,
	             name_manual: str,
	             config: habapp_rules.actors.config.ventilation.VentilationConfig,
	             name_hand_request: str | None = None,
	             name_external_request: str | None = None,
	             name_presence_state: str | None = None,
	             name_feedback_on: str | None = None,
	             name_feedback_power: str | None = None,
	             name_display_text: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of ventilation base.

		:param name_manual: name of OpenHAB switch item to disable all automatic functions  (SwitchItem)
		:param config: configuration of the ventilation
		:param name_hand_request: name of OpenHAB switch item to enter the hand state (SwitchItem)
		:param name_external_request: name of OpenHAB switch item to enter the external state (e.g. used for dryer request) (SwitchItem)
		:param name_presence_state: name of OpenHAB presence state item (StringItem)
		:param name_feedback_on: name of OpenHAB item which shows that ventilation is on (SwitchItem)
		:param name_feedback_power: name of OpenHAB item which shows that ventilation is in power mode (SwitchItem)
		:param name_display_text: name of OpenHAB item which can be used to set the display text (e.g. for user interface) (StringItem)
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""

		self._config = config
		self._ventilation_level: int | None = None

		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, name_state, state_label)

		# init items
		self._item_manual = HABApp.openhab.items.SwitchItem.get_item(name_manual)
		self._item_hand_request = HABApp.openhab.items.SwitchItem.get_item(name_hand_request) if name_hand_request else None
		self._item_external_request = HABApp.openhab.items.SwitchItem.get_item(name_external_request) if name_external_request else None
		self._item_presence_state = HABApp.openhab.items.StringItem.get_item(name_presence_state) if name_presence_state else None
		self._item_feedback_on = HABApp.openhab.items.SwitchItem.get_item(name_feedback_on) if name_feedback_on else None
		self._item_feedback_power = HABApp.openhab.items.SwitchItem.get_item(name_feedback_power) if name_feedback_power else None
		self._item_display_text = HABApp.openhab.items.StringItem.get_item(name_display_text) if name_display_text else None

		# init state machine
		self._previous_state = None
		self._state_change_time = datetime.datetime.now()
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self._apply_config()

		# callbacks
		self._item_manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_hand_request is not None:
			self._item_hand_request.listen_event(self._cb_power_hand_request, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_external_request is not None:
			self._item_external_request.listen_event(self._cb_external_request, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_presence_state is not None:
			self._item_presence_state.listen_event(self._cb_presence_state, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._update_openhab_state()

	def _get_initial_state(self, default_value: str = "initial") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_manual.is_on():
			return "Manual"
		if self._item_hand_request is not None and self._item_hand_request.is_on():
			return "Auto_PowerHand"
		if self._item_presence_state is not None and self._item_presence_state.value == habapp_rules.system.PresenceState.LONG_ABSENCE.value:
			return "Auto_LongAbsence"
		if self._item_external_request is not None and self._item_external_request.is_on():
			return "Auto_PowerExternal"
		return "Auto_Normal"

	def _apply_config(self) -> None:
		"""Apply values from config."""
		self.state_machine.get_state("Auto_PowerHand").timeout = self._config.state_hand.timeout
		self.state_machine.get_state("Auto_LongAbsence_On").timeout = self._config.state_long_absence.duration

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item and other states.

		This method should be set to "after_state_change" of the state machine.
		"""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self._state_change_time = datetime.datetime.now()
			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

			self._set_level()
			self._set_feedback_states()
			self._previous_state = self.state

	def _set_level(self) -> None:
		"""Set ventilation level"""
		if self.state == "Manual":
			return

		if self.state == "Auto_PowerHand":
			self._ventilation_level = self._config.state_hand.level
		elif self.state == "Auto_Normal":
			self._ventilation_level = self._config.state_normal.level
		elif self.state == "Auto_PowerExternal":
			self._ventilation_level = self._config.state_external.level
		elif self.state == "Auto_LongAbsence_On":
			self._ventilation_level = self._config.state_long_absence.level
		elif self.state == "Auto_LongAbsence_Off":
			self._ventilation_level = 0
		else:
			return

		self._set_level_to_ventilation_items()

	@abc.abstractmethod
	def _set_level_to_ventilation_items(self) -> None:
		"""Set ventilation to output item(s)."""

	def _get_display_text(self) -> str | None:
		"""Get Text for display.

		:return: text for display or None if not defined for this state
		"""
		if self.state == "Manual":
			return "Manual"
		if self.state == "Auto_Normal":
			return self._config.state_normal.display_text
		if self.state == "Auto_PowerExternal":
			return self._config.state_external.display_text
		if self.state == "Auto_LongAbsence_On":
			return f"{self._config.state_long_absence.display_text} ON"
		if self.state == "Auto_LongAbsence_Off":
			return f"{self._config.state_long_absence.display_text} OFF"

		return None

	def _set_feedback_states(self) -> None:
		"""Set feedback sates to the OpenHAB items."""
		if self._item_hand_request is not None and self._previous_state == "Auto_PowerHand":
			habapp_rules.core.helper.send_if_different(self._item_hand_request, "OFF")

		if self._item_feedback_on is not None:
			habapp_rules.core.helper.send_if_different(self._item_feedback_on, "ON" if self._ventilation_level else "OFF")

		if self._item_feedback_power is not None:
			target_value = "ON" if self._ventilation_level is not None and self._ventilation_level >= 2 else "OFF"
			habapp_rules.core.helper.send_if_different(self._item_feedback_power, target_value)

		if self._item_display_text is not None:
			if self.state == "Auto_PowerHand":
				self.__set_hand_display_text()
				return

			if (display_text := self._get_display_text()) is not None:
				habapp_rules.core.helper.send_if_different(self._item_display_text, display_text)

	def __set_hand_display_text(self) -> None:
		"""Callback to set display text."""
		if self.state != "Auto_PowerHand":
			# state changed and is not PowerHand anymore
			return

		# get the remaining minutes and set display text
		remaining_minutes = round((self._config.state_hand.timeout - (datetime.datetime.now() - self._state_change_time).seconds) / 60)
		remaining_minutes = remaining_minutes if remaining_minutes >= 0 else 0
		habapp_rules.core.helper.send_if_different(self._item_display_text, f"{self._config.state_hand.display_text} {remaining_minutes}min")

		# re-trigger this method in 1 minute
		self.run.at(60, self.__set_hand_display_text)

	def on_enter_Auto_Init(self) -> None:  # pylint: disable=invalid-name
		"""Is called on entering of Auto_Init state"""
		self._set_initial_state()

	def on_enter_Auto_LongAbsence_Off(self):  # pylint: disable=invalid-name
		"""Is called on entering of Auto_LongAbsence_Off state."""
		self.run.at(self._config.state_long_absence.start_time, self._long_absence_power_on)

	def _external_active_and_configured(self) -> bool:
		"""Check if external request is active and configured.

		:return: True if external request is active and configured
		"""
		return self._item_external_request is not None and self._item_external_request.is_on()

	def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if manual mode changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._manual_on()
		else:
			self._manual_off()

	def _cb_power_hand_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if power_hand_request changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._hand_on()
		else:
			self._hand_off()

	def _cb_external_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if external request changed.

		:param event: original trigger event
		"""
		if event.value == "ON":
			self._external_on()
		else:
			self._external_off()

	def _cb_presence_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if presence_state changed.

		:param event: original trigger event
		"""
		if event.value == habapp_rules.system.PresenceState.LONG_ABSENCE.value:
			self._long_absence_on()
		else:
			self._long_absence_off()


class Ventilation(_VentilationBase):
	"""Rule for managing ventilation systems which can be controlled with ventilation levels

	# Items:
	Number  Ventilation_level           "Ventilation level"
	Switch  Manual                      "Manual"
	Switch  Hand_Request                "Hand request"
	Switch  External_Request            "External request"
	String  presence_state              "Presence state"
	Switch  Feedback_On                 "Feedback is ON"
	Switch  Feedback_Power              "Feedback is Power"

	# Rule init:
	habapp_rules.actors.ventilation.Ventilation(
		"Ventilation_level",
		"Manual",
		habapp_rules.actors.config.ventilation.CONFIG_DEFAULT,
		"Hand_Request",
		name_external_request="External_Request",
		name_presence_state="presence_state",
		name_feedback_on="Feedback_On",
		name_feedback_power="Feedback_Power"
	)
	"""

	# pylint: disable=too-many-arguments
	def __init__(self,
	             name_ventilation_level: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.ventilation.VentilationConfig,
	             name_hand_request: str | None = None,
	             name_external_request: str | None = None,
	             name_presence_state: str | None = None,
	             name_feedback_on: str | None = None,
	             name_feedback_power: str | None = None,
	             name_display_text: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of ventilation object.

		:param name_ventilation_level: name of OpenHAB number item to set the ventilation level (NumberItem
		:param name_manual: name of OpenHAB switch item to disable all automatic functions  (SwitchItem)
		:param config: configuration of the ventilation
		:param name_hand_request: name of OpenHAB switch item to enter the hand state (SwitchItem)
		:param name_external_request: name of OpenHAB switch item to enter the external state (e.g. used for dryer request) (SwitchItem)
		:param name_presence_state: name of OpenHAB presence state item (StringItem)
		:param name_feedback_on: name of OpenHAB item which shows that ventilation is on (SwitchItem)
		:param name_feedback_power: name of OpenHAB item which shows that ventilation is in power mode (SwitchItem)
		:param name_display_text: name of OpenHAB item which can be used to set the display text (e.g. for user interface) (StringItem)
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		if not name_state:
			name_state = f"H_{name_ventilation_level}_state"
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_ventilation_level)

		self._item_ventilation_level = HABApp.openhab.items.NumberItem.get_item(name_ventilation_level)

		_VentilationBase.__init__(
			self,
			name_manual,
			config,
			name_hand_request,
			name_external_request,
			name_presence_state,
			name_feedback_on,
			name_feedback_power,
			name_display_text,
			name_state,
			state_label
		)
		self._instance_logger.info(habapp_rules.core.state_machine_rule.StateMachineRule.get_initial_log_message(self))

	def _set_level_to_ventilation_items(self) -> None:
		"""Set ventilation to output item(s)."""
		habapp_rules.core.helper.send_if_different(self._item_ventilation_level, self._ventilation_level)


class VentilationHeliosTwoStage(_VentilationBase):
	"""Rule for managing Helios ventilation systems with humidity sensor (E.g. Helios ELS)

	# Items:
	Switch  Ventilation_Switch_On       "Ventilation relay on"
	Switch  Ventilation_Switch_Power    "Ventilation relay power"
	Switch  Manual                      "Manual"
	Switch  Hand_Request                "Hand request"
	Switch  External_Request            "External request"
	String  presence_state              "Presence state"
	Switch  Feedback_On                 "Feedback is ON"
	Switch  Feedback_Power              "Feedback is Power"

	# Rule init:
	habapp_rules.actors.ventilation.VentilationHeliosTwoStage(
		"Ventilation_Switch_On",
		"Ventilation_Switch_Power",
		"Manual",
		habapp_rules.actors.config.ventilation.CONFIG_DEFAULT,
		"Hand_Request",
		name_external_request="External_Request",
		name_presence_state="presence_state",
		name_feedback_on="Feedback_On",
		name_feedback_power="Feedback_Power"
	)
	"""
	states = copy.deepcopy(_VentilationBase.states)
	__AUTO_STATE = next(state for state in states if state["name"] == "Auto")  # pragma: no cover
	__AUTO_STATE["children"].append({"name": "PowerAfterRun", "timeout": 390, "on_timeout": "_after_run_timeout"})

	trans = copy.deepcopy(_VentilationBase.trans)
	# remove not needed transitions
	trans.remove({"trigger": "_hand_on", "source": ["Auto_Normal", "Auto_PowerExternal", "Auto_LongAbsence"], "dest": "Auto_PowerHand"})  # will be extended with additional source state
	trans.remove({"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_Normal", "unless": "_external_active_and_configured"})  # this is not needed anymore since there is always Auto_PowerAfterRun after any power state
	trans.remove({"trigger": "_external_on", "source": "Auto_Normal", "dest": "Auto_PowerExternal"})  # will be extended with additional source state
	trans.remove({"trigger": "_external_off", "source": "Auto_PowerExternal", "dest": "Auto_Normal"})  # this is not needed anymore since there is always Auto_PowerAfterRun after any power state

	# add new PowerHand transitions
	trans.append({"trigger": "_hand_on", "source": ["Auto_Normal", "Auto_PowerExternal", "Auto_LongAbsence", "Auto_PowerAfterRun"], "dest": "Auto_PowerHand"})
	trans.append({"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_PowerAfterRun", "unless": "_external_active_and_configured"})

	# add new PowerExternal transitions
	trans.append({"trigger": "_external_on", "source": ["Auto_Normal", "Auto_PowerAfterRun"], "dest": "Auto_PowerExternal"})
	trans.append({"trigger": "_external_off", "source": "Auto_PowerExternal", "dest": "Auto_PowerAfterRun"})

	# add new PowerAfterRun transitions
	trans.append({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal"})

	# pylint: disable=too-many-arguments
	def __init__(self,
	             name_ventilation_output_on: str,
	             name_ventilation_output_power: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.ventilation.VentilationConfig,
	             name_hand_request: str | None = None,
	             name_external_request: str | None = None,
	             name_presence_state: str | None = None,
	             name_feedback_on: str | None = None,
	             name_feedback_power: str | None = None,
	             name_display_text: str | None = None,
	             after_run_timeout: int = 390,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of a Helios ventilation object which uses two switches to set the level.

		:param name_ventilation_output_on: name of OpenHAB switch item to switch on the ventilation (SwitchItem)
		:param name_ventilation_output_power: name of OpenHAB switch item to switch on the power mode (SwitchItem)
		:param name_manual: name of OpenHAB switch item to disable all automatic functions  (SwitchItem)
		:param config: configuration of the ventilation
		:param name_hand_request: name of OpenHAB switch item to enter the hand state (SwitchItem)
		:param name_external_request: name of OpenHAB switch item to enter the external state (e.g. used for dryer request) (SwitchItem)
		:param name_presence_state: name of OpenHAB presence state item (StringItem)
		:param name_feedback_on: name of OpenHAB item which shows that ventilation is on (SwitchItem)
		:param name_feedback_power: name of OpenHAB item which shows that ventilation is in power mode (SwitchItem)
		:param name_display_text: name of OpenHAB item which can be used to set the display text (e.g. for user interface) (StringItem)
		:param after_run_timeout: timeout of after-run-state in seconds
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""

		if not name_state:
			name_state = f"H_{name_ventilation_output_on}_state"
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_ventilation_output_on)

		# get items
		self._item_ventilation_on = HABApp.openhab.items.SwitchItem.get_item(name_ventilation_output_on)
		self._item_ventilation_power = HABApp.openhab.items.SwitchItem.get_item(name_ventilation_output_power)

		_VentilationBase.__init__(
			self,
			name_manual,
			config,
			name_hand_request,
			name_external_request,
			name_presence_state,
			name_feedback_on,
			name_feedback_power,
			name_display_text,
			name_state,
			state_label
		)

		# set timeout
		self.state_machine.get_state("Auto_PowerAfterRun").timeout = after_run_timeout

		self._instance_logger.info(habapp_rules.core.state_machine_rule.StateMachineRule.get_initial_log_message(self))

	def _get_display_text(self) -> str | None:
		"""Get Text for display.

		:return: text for display or None if not defined for this state
		"""
		if self.state == "Auto_PowerAfterRun":
			return self._config.state_after_run.display_text

		return _VentilationBase._get_display_text(self)

	def _set_level(self) -> None:
		if self.state == "Auto_PowerAfterRun":
			self._ventilation_level = self._config.state_after_run.level
			self._set_level_to_ventilation_items()
			return

		super()._set_level()

	def _set_level_to_ventilation_items(self) -> None:
		"""Set ventilation to output item(s)."""
		if self.state == "Auto_PowerAfterRun":
			habapp_rules.core.helper.send_if_different(self._item_ventilation_on, "ON")
			habapp_rules.core.helper.send_if_different(self._item_ventilation_power, "OFF")

		else:
			habapp_rules.core.helper.send_if_different(self._item_ventilation_on, "ON" if self._ventilation_level else "OFF")
			habapp_rules.core.helper.send_if_different(self._item_ventilation_power, "ON" if self._ventilation_level >= 2 else "OFF")


# pylint: disable=no-member, missing-return-doc
class VentilationHeliosTwoStageHumidity(VentilationHeliosTwoStage):
	"""Rule for managing Helios ventilation systems with humidity sensor (E.g. Helios ELS)

	# Items:
	Switch  Ventilation_Switch_On       "Ventilation relay on"
	Switch  Ventilation_Switch_Power    "Ventilation relay power"
	Number  Ventilation_Current         "Ventilation current"
	Switch  Manual                      "Manual"
	Switch  Hand_Request                "Hand request"
	Switch  External_Request            "External request"
	String  presence_state              "Presence state"
	Switch  Feedback_On                 "Feedback is ON"
	Switch  Feedback_Power              "Feedback is Power"

	# Rule init:
	habapp_rules.actors.ventilation.VentilationHeliosTwoStageHumidity(
		"Ventilation_Switch_On",
		"Ventilation_Switch_Power",
		"Ventilation_Current",
		"Manual",
		habapp_rules.actors.config.ventilation.CONFIG_DEFAULT,
		"Hand_Request",
		name_external_request="External_Request",
		name_presence_state="presence_state",
		name_feedback_on="Feedback_On",
		name_feedback_power="Feedback_Power"
	)
	"""
	states = copy.deepcopy(VentilationHeliosTwoStage.states)
	__AUTO_STATE = next(state for state in states if state["name"] == "Auto")  # pragma: no cover
	__AUTO_STATE["children"].append({"name": "PowerHumidity"})

	trans = copy.deepcopy(VentilationHeliosTwoStage.trans)
	# remove not needed transitions
	trans.remove({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal"})  # will be changed to only go to AutoNormal if the current is below the threshold (not humidity)

	# add new PowerHumidity transitions
	trans.append({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal", "unless": "_current_greater_threshold"})
	trans.append({"trigger": "_end_after_run", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal"})
	trans.append({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_PowerHumidity", "conditions": "_current_greater_threshold"})

	trans.append({"trigger": "_humidity_on", "source": "Auto_Normal", "dest": "Auto_PowerHumidity"})
	trans.append({"trigger": "_humidity_off", "source": "Auto_PowerHumidity", "dest": "Auto_Normal"})

	trans.append({"trigger": "_hand_on", "source": "Auto_PowerHumidity", "dest": "Auto_PowerHand"})
	trans.append({"trigger": "_external_on", "source": "Auto_PowerHumidity", "dest": "Auto_PowerExternal"})

	# pylint: disable=too-many-locals, too-many-arguments
	def __init__(self,
	             name_ventilation_output_on: str,
	             name_ventilation_output_power: str,
	             name_current: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.ventilation.VentilationConfig,
	             name_hand_request: str | None = None,
	             name_external_request: str | None = None,
	             name_presence_state: str | None = None,
	             name_feedback_on: str | None = None,
	             name_feedback_power: str | None = None,
	             name_display_text: str | None = None,
	             after_run_timeout: int = 390,
	             current_threshold_power: float = 0.105,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of a Helios ventilation object which uses two switches to set the level, including a humidity sensor.

		:param name_ventilation_output_on: name of OpenHAB switch item to switch on the ventilation (SwitchItem)
		:param name_ventilation_output_power: name of OpenHAB switch item to switch on the power mode (SwitchItem)
		:param name_current: name of OpenHAB number item which measures the current of the ventilation (NumberItem)
		:param name_manual: name of OpenHAB switch item to disable all automatic functions  (SwitchItem)
		:param config: configuration of the ventilation
		:param name_hand_request: name of OpenHAB switch item to enter the hand state (SwitchItem)
		:param name_external_request: name of OpenHAB switch item to enter the external state (e.g. used for dryer request) (SwitchItem)
		:param name_presence_state: name of OpenHAB presence state item (StringItem)
		:param name_feedback_on: name of OpenHAB item which shows that ventilation is on (SwitchItem)
		:param name_feedback_power: name of OpenHAB item which shows that ventilation is in power mode (SwitchItem)
		:param name_display_text: name of OpenHAB item which can be used to set the display text (e.g. for user interface) (StringItem)
		:param after_run_timeout: timeout of after-run-state in seconds
		:param current_threshold_power: Threshold of the current which is used in power mode (NumberItem)
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		self._item_current = HABApp.openhab.items.NumberItem.get_item(name_current)
		self._current_threshold_power = current_threshold_power

		VentilationHeliosTwoStage.__init__(
			self,
			name_ventilation_output_on,
			name_ventilation_output_power,
			name_manual,
			config,
			name_hand_request,
			name_external_request,
			name_presence_state,
			name_feedback_on,
			name_feedback_power,
			name_display_text,
			after_run_timeout,
			name_state,
			state_label
		)

		self._item_current.listen_event(self._cb_current, HABApp.openhab.events.ItemStateUpdatedEventFilter())

	def _get_initial_state(self, default_value: str = "initial") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		state = super()._get_initial_state(default_value)

		if state == "Auto_Normal" and self._current_greater_threshold():
			return "Auto_PowerHumidity"

		return state

	def _get_display_text(self) -> str | None:
		"""Get Text for display.

		:return: text for display or None if not defined for this state
		"""
		if self.state == "Auto_PowerHumidity":
			return self._config.state_humidity.display_text

		return VentilationHeliosTwoStage._get_display_text(self)

	def _set_level(self) -> None:
		if self.state == "Auto_PowerHumidity":
			self._ventilation_level = self._config.state_humidity.level
			self._set_level_to_ventilation_items()
			return

		super()._set_level()

	def _set_level_to_ventilation_items(self) -> None:
		"""Set ventilation to output item(s)."""
		if self.state == "Auto_PowerHumidity":
			habapp_rules.core.helper.send_if_different(self._item_ventilation_on, "ON")
			habapp_rules.core.helper.send_if_different(self._item_ventilation_power, "OFF")
		else:
			super()._set_level_to_ventilation_items()

	def _current_greater_threshold(self, current: float | None = None) -> bool:
		"""Check if current is greater than the threshold

		:param current: current which should be checked. If None the value of the current item will be taken
		:return: True if current greater than the threshold, else False
		"""
		current = current if current is not None else self._item_current.value

		if current is None:
			return False

		return current > self._current_threshold_power

	def _cb_current(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
		"""Callback which is triggered if the current changed.

		:param event: original trigger event
		"""
		if self.state != "Auto_PowerHumidity" and self._current_greater_threshold(event.value):
			self._humidity_on()
		elif self.state == "Auto_PowerHumidity" and not self._current_greater_threshold(event.value):
			self._humidity_off()
		elif self.state == "Auto_PowerAfterRun" and not self._current_greater_threshold(event.value):
			self._end_after_run()



"""Rules for astro actions."""
import abc
import logging

import HABApp

import habapp_rules.core.helper

LOGGER = logging.getLogger(__name__)


class _SetNightDayBase(HABApp.Rule):
	"""Base class for set night / day."""

	def __init__(self, name_target: str, name_elevation: str, elevation_threshold: float) -> None:
		"""Init Rule.

		:param name_target: name of OpenHAB switch item which should be set depending on the sun elevation value
		:param name_elevation: name of OpenHAB elevation (NumberItem)
		:param elevation_threshold: Threshold value for elevation.
		"""
		HABApp.Rule.__init__(self)

		self._item_target = HABApp.openhab.items.SwitchItem.get_item(name_target)
		self._item_elevation = HABApp.openhab.items.NumberItem.get_item(name_elevation)
		self._elevation_threshold = elevation_threshold

		self._item_elevation.listen_event(self._set_night, HABApp.openhab.events.ItemStateChangedEventFilter())

		self.run.soon(self._set_night)

	def _set_night(self, _: HABApp.openhab.events.ItemStateChangedEvent | None = None) -> None:
		"""Callback which sets the state to the night item."""
		if self._item_elevation.value is None:
			return
		habapp_rules.core.helper.send_if_different(self._item_target, self._get_target_value())

	@abc.abstractmethod
	def _get_target_value(self) -> str:
		"""Get target value which should be set.

		:return: target value (ON / OFF)
		"""


class SetDay(_SetNightDayBase):
	"""Rule to set / unset day item at dusk / dawn.

		# Items:
		Switch    night_for_shading     "Night for shading"
		Number    elevation             "Sun elevation [%s]"    <sun>     {channel="astro...}

		# Rule init:
		habapp_rules.actors.shading.SetNight("night_for_shading", "elevation")
	"""

	def __init__(self, name_day: str, name_elevation: str, elevation_threshold: float = 0) -> None:
		"""Init Rule.

		:param name_day: name of OpenHAB switch item which should be set to "ON" after dawn and "OFF after dusk
		:param name_elevation: name of OpenHAB elevation (NumberItem)
		:param elevation_threshold: Threshold value for elevation. If the sun elevation is greater than the threshold, the day item will be set to "ON"
		"""
		_SetNightDayBase.__init__(self, name_day, name_elevation, elevation_threshold)

	def _get_target_value(self) -> str:
		"""Get target value which should be set.

		:return: target value (ON / OFF)
		"""
		return "ON" if self._item_elevation.value > self._elevation_threshold else "OFF"


class SetNight(_SetNightDayBase):
	"""Rule to set / unset night item at dusk / dawn.

		# Items:
		Switch    night_for_shading     "Night for shading"
		Number    elevation             "Sun elevation [%s]"    <sun>     {channel="astro...}

		# Rule init:
		habapp_rules.actors.shading.SetNight("night_for_shading", "elevation")
	"""

	def __init__(self, name_night: str, name_elevation: str, elevation_threshold: float = -8) -> None:
		"""Init Rule.

		:param name_night: name of OpenHAB switch item which should be set to "ON" after dusk and "OFF after dawn
		:param name_elevation: name of OpenHAB elevation (NumberItem)
		:param elevation_threshold: Threshold value for elevation. If the sun elevation is below the threshold, the night item will be set to "ON"
		"""
		_SetNightDayBase.__init__(self, name_night, name_elevation, elevation_threshold)

	def _get_target_value(self) -> str:
		"""Get target value which should be set.

		:return: target value (ON / OFF)
		"""
		return "ON" if self._item_elevation.value < self._elevation_threshold else "OFF"



"""Set version rules."""
import logging

import HABApp

import habapp_rules.__version__
import habapp_rules.core.helper

LOGGER = logging.getLogger(__name__)


class SetVersions(HABApp.Rule):
	"""Update HABApp and habapp_rules version to OpenHAB items."""

	def __init__(self) -> None:
		"""init rule."""
		HABApp.Rule.__init__(self)
		LOGGER.info("Update versions of OpenHAB items")

		item_version_habapp = habapp_rules.core.helper.create_additional_item("H_habapp_version", "String", "HABApp version")
		item_version_habapp_rules = habapp_rules.core.helper.create_additional_item("H_habapp_rules_version", "String", "habapp_rules version")

		item_version_habapp.oh_send_command(HABApp.__version__)
		item_version_habapp_rules.oh_send_command(habapp_rules.__version__.__version__)



"""Implementations for observing states of switch / dimmer / roller shutter."""
from __future__ import annotations

import abc
import logging
import threading
import time
import typing

import HABApp
import HABApp.openhab.items

import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.timeout_list

LOGGER = logging.getLogger(__name__)

EventTypes = typing.Union[HABApp.openhab.events.ItemStateChangedEvent, HABApp.openhab.events.ItemCommandEvent]
CallbackType = typing.Callable[[EventTypes], None]


class _StateObserverBase(HABApp.Rule, abc.ABC):
	"""Base class for observer classes."""

	def __init__(self, item_name: str, control_names: list[str] | None = None, group_names: list[str] | None = None, value_tolerance: float = 0):
		"""Init state observer for switch item.

		:param item_name: Name of observed item
		:param control_names: list of control items.
		:param group_names: list of group items where the item is a part of. Group item type must match with type of item_name
		:param value_tolerance: used by all observers which handle numbers. It can be used to allow a difference when comparing new and old values.
		"""
		self._value_tolerance = value_tolerance

		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, item_name)

		self._last_manual_event = HABApp.openhab.events.ItemCommandEvent("", None)

		self._item = HABApp.openhab.items.OpenhabItem.get_item(item_name)

		self.__control_items = [HABApp.openhab.items.OpenhabItem.get_item(name) for name in control_names] if control_names else []
		self.__group_items = [HABApp.openhab.items.OpenhabItem.get_item(name) for name in group_names] if group_names else []
		self.__check_item_types()

		self._value = self._item.value
		self._group_last_event = 0

		self._item.listen_event(self._cb_item, HABApp.openhab.events.ItemStateChangedEventFilter())
		for control_item in self.__control_items:
			control_item.listen_event(self._cb_control_item, HABApp.openhab.events.ItemCommandEventFilter())
		for group_item in self.__group_items:
			group_item.listen_event(self._cb_group_item, HABApp.openhab.events.ItemStateUpdatedEventFilter())

	@property
	def value(self) -> float | bool:
		"""Get the current state / value of the observed item."""
		return self._value

	@property
	def last_manual_event(self) -> EventTypes:
		"""Get the last manual event."""
		return self._last_manual_event

	def __check_item_types(self) -> None:
		"""Check if all command and control items have the correct type.

		:raises TypeError: if one item has the wrong type"""
		target_type = type(self._item)

		wrong_types = []
		for item in self.__control_items + self.__group_items:
			if not isinstance(item, target_type):
				wrong_types.append(f"{item.name} <{type(item).__name__}>")

		if wrong_types:
			self._instance_logger.error(msg := f"Found items with wrong item type. Expected: {target_type.__name__}. Wrong: {' | '.join(wrong_types)}")
			raise TypeError(msg)

	@abc.abstractmethod
	def send_command(self, value: float | str) -> None:
		"""Send brightness command to light (this should be used by rules, to not trigger a manual action)

		:param value: Value to send to the light
		:raises ValueError: if value has wrong format
		"""

	def _cb_item(self, event: HABApp.openhab.events.ItemStateChangedEvent):
		"""Callback, which is called if a value change of the light item was detected.

		:param event: event, which triggered this callback
		"""
		self._check_manual(event)

	def _cb_group_item(self, event: HABApp.openhab.events.ItemStateUpdatedEvent):
		"""Callback, which is called if a value change of the light item was detected.

		:param event: event, which triggered this callback
		"""
		if event.value in {"ON", "OFF"} and time.time() - self._group_last_event > 0.3:  # this is some kind of workaround. For some reason all events are doubled.
			self._group_last_event = time.time()
			self._check_manual(event)

	@abc.abstractmethod
	def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent):
		"""Callback, which is called if a command event of one of the control items was detected.

		:param event: event, which triggered this callback
		"""

	@abc.abstractmethod
	def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Check if light was triggered by a manual action

		:param event: event which triggered this method. This will be forwarded to the callback
		:raises ValueError: if event is not supported
		"""

	def _trigger_callback(self, cb_name: str, event: EventTypes) -> None:
		"""Trigger a manual detected callback.

		:param cb_name: name of callback method
		:param event: event which triggered the callback
		"""
		self._last_manual_event = event
		callback: CallbackType = getattr(self, cb_name)
		callback(event)

	def _values_different_with_tolerance(self, value_1: float, value_2: float) -> bool:
		"""Check if values are different, including the difference.

		:param value_1: first value
		:param value_2: second value
		:return: true if values are different (including the tolerance), false if not
		"""
		return abs((value_1 or 0) - (value_2 or 0)) > self._value_tolerance


class StateObserverSwitch(_StateObserverBase):
	"""Class to observe the on/off state of a switch item.

	This class is normally not used standalone. Anyway here is an example config:

	# KNX-things:
	Thing device T00_99_OpenHab_DimmerSwitch "KNX OpenHAB switch observer"{
        Type switch             : switch             "Switch"             [ switch="1/1/10" ]
    }

    # Items:
	    Switch    I01_01_Switch    "Switch [%s]" {channel="knx:device:bridge:T00_99_OpenHab_DimmerSwitch:switch"}

	# Rule init:
	habapp_rules.actors.state_observer.StateObserverSwitch("I01_01_Switch", callback_on, callback_off)
	"""

	def __init__(self, item_name: str, cb_on: CallbackType, cb_off: CallbackType):
		"""Init state observer for switch item.

		:param item_name: Name of switch item
		:param cb_on: callback which should be called if manual_on was detected
		:param cb_off: callback which should be called if manual_off was detected
		"""
		self._cb_on = cb_on
		self._cb_off = cb_off
		_StateObserverBase.__init__(self, item_name)
		self._value = self._item.value

	def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Check if light was triggered by a manual action

		:param event: event which triggered this method. This will be forwarded to the callback
		:raises ValueError: if event is not supported
		"""
		if event.value == "ON" and not self._value:
			self._value = True
			self._trigger_callback("_cb_on", event)

		elif event.value == "OFF" and self._value:
			self._value = False
			self._trigger_callback("_cb_off", event)

	def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent):  # not used by StateObserverSwitch
		"""Callback, which is called if a command event of one of the control items was detected.

		:param event: event, which triggered this callback
		"""

	def send_command(self, value: str) -> None:
		"""Send brightness command to light (this should be used by rules, to not trigger a manual action)

		:param value: Value to send to the light
		:raises ValueError: if value has wrong format
		"""
		if value == "ON":
			self._value = True

		elif value == "OFF":
			self._value = False
		else:
			raise ValueError(f"The given value is not supported for StateObserverSwitch: {value}")

		self._item.oh_send_command(value)


class StateObserverDimmer(_StateObserverBase):
	"""Class to observe the on / off / change events of a dimmer item.

	Known limitation: if the items of group_names are KNX-items, the channel types must be dimmer (not dimmer-control)
	This class is normally not used standalone. Anyway here is an example config:

	# KNX-things:
	Thing device T00_99_OpenHab_DimmerObserver "KNX OpenHAB dimmer observer"{
        Type dimmer             : light             "Light"             [ switch="1/1/10", position="1/1/13+<1/1/15" ]
        Type dimmer-control     : light_ctr         "Light control"     [ increaseDecrease="1/1/12"]
        Type dimmer             : light_group       "Light Group"       [ switch="1/1/240", position="1/1/243"]
    }

    # Items:
    Dimmer    I01_01_Light              "Light [%s]"        {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light"}
	Dimmer    I01_01_Light_ctr          "Light ctr"         {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_ctr"}
	Dimmer    I01_01_Light_group        "Light Group"       {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_group"}

	# Rule init:
	habapp_rules.actors.state_observer.StateObserverDimmer(
			"I01_01_Light",
			control_names=["I01_01_Light_ctr"],
			group_names=["I01_01_Light_group"],
			cb_on=callback_on,
			cb_off=callback_off,
			cb_brightness_change=callback_change)
	"""

	def __init__(self, item_name: str, cb_on: CallbackType, cb_off: CallbackType, cb_brightness_change: CallbackType | None = None, control_names: list[str] | None = None, group_names: list[str] | None = None, value_tolerance: float = 0) -> None:
		"""Init state observer for dimmer item.

		:param item_name: Name of dimmer item
		:param cb_on: callback which is called if manual_on was detected
		:param cb_off: callback which is called if manual_off was detected
		:param cb_brightness_change: callback which is called if dimmer is on and brightness changed
		:param control_names: list of control items. They are used to also respond to switch on/off via INCREASE/DECREASE
		:param group_names: list of group items where the item is a part of. Group item type must match with type of item_name
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		"""

		_StateObserverBase.__init__(self, item_name, control_names, group_names, value_tolerance)

		self._cb_on = cb_on
		self._cb_off = cb_off
		self._cb_brightness_change = cb_brightness_change

	def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Check if light was triggered by a manual action

		:param event: event which triggered this method. This will be forwarded to the callback
		"""
		if isinstance(event.value, (int, float)):
			if event.value > 0 and (self._value is None or self._value == 0):
				self._value = event.value
				self._trigger_callback("_cb_on", event)

			elif event.value == 0 and (self._value is None or self._value > 0):
				self._value = 0
				self._trigger_callback("_cb_off", event)

			elif self._values_different_with_tolerance(event.value, self._value):
				self._value = event.value
				self._trigger_callback("_cb_brightness_change", event)

		elif event.value == "ON" and (self._value is None or self._value == 0):
			self._value = 100
			self._trigger_callback("_cb_on", event)

		elif event.value == "OFF" and (self._value is None or self._value > 0):
			self._value = 0
			self._trigger_callback("_cb_off", event)

	def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent):
		"""Callback, which is called if a command event of one of the control items was detected.

		:param event: event, which triggered this callback
		"""
		if event.value == "INCREASE" and (self._value is None or self._value == 0):
			self._value = 100
			self._trigger_callback("_cb_on", event)

	def send_command(self, value: float | str) -> None:
		"""Send brightness command to light (this should be used by rules, to not trigger a manual action)

		:param value: Value to send to the light
		:raises ValueError: if value has wrong format
		"""
		if isinstance(value, (int, float)):
			self._value = value

		elif value == "ON":
			self._value = 100

		elif value == "OFF":
			self._value = 0

		else:
			raise ValueError(f"The given value is not supported for StateObserverDimmer: {value}")

		self._item.oh_send_command(value)


class StateObserverRollerShutter(_StateObserverBase):
	"""Class to observe manual controls of a roller shutter item.

		This class is normally not used standalone. Anyway, here is an example config:

		# KNX-things:
		Thing device T00_99_OpenHab_RollershutterObserver "KNX OpenHAB rollershutter observer"{
	        Type rollershutter             : shading             "Shading"             [ upDown="1/1/10", position="1/1/13+<1/1/15" ]
	        Type rollershutter-control     : shading_ctr         "Shading control"     [ upDown="1/1/10", position="1/1/13+<1/1/15" ]
	        Type rollershutter             : shading_group       "Shading Group"       [ upDown="1/1/110", position="1/1/113+<1/1/115" ]
	    }

	    # Items:
	    Rollershutter    I_Rollershutter              "Rollershutter [%s]"        {channel="knx:device:bridge:T00_99_OpenHab_RollershutterObserver:Rollershutter"}
		Rollershutter    I_Rollershutter_ctr          "Rollershutter ctr"         {channel="knx:device:bridge:T00_99_OpenHab_RollershutterObserver:Rollershutter_ctr"}
		Rollershutter    I_Rollershutter_group        "Rollershutter Group"       {channel="knx:device:bridge:T00_99_OpenHab_RollershutterObserver:Rollershutter_group"}

		# Rule init:
		habapp_rules.actors.state_observer.StateObserverRollerShutter(
				"I_Rollershutter",
				control_names=["I_Rollershutter_ctr"],
				group_names=["I_Rollershutter_group"],
				cb_manual=callback_on
				)
		"""

	def __init__(self, item_name: str, cb_manual: CallbackType, control_names: list[str] | None = None, group_names: list[str] | None = None, value_tolerance: float = 0) -> None:
		"""Init state observer for dimmer item.

		:param item_name: Name of dimmer item
		:param cb_manual: callback which is called if a manual interaction was detected
		:param control_names: list of control items. They are used to also respond to switch on/off via INCREASE/DECREASE
		:param group_names: list of group items where the item is a part of. Group item type must match with type of item_name
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		"""
		self._value_tolerance = value_tolerance
		_StateObserverBase.__init__(self, item_name, control_names, group_names, value_tolerance)

		self._cb_manual = cb_manual

	def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		if isinstance(event.value, (int, float)) and self._values_different_with_tolerance(event.value, self._value):
			self._value = event.value
			self._trigger_callback("_cb_manual", event)

	def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent):
		if event.value == "DOWN":
			self._value = 100
			self._trigger_callback("_cb_manual", event)

		elif event.value == "UP":
			self._value = 0
			self._trigger_callback("_cb_manual", event)

	def send_command(self, value: float) -> None:
		if not isinstance(value, (int, float)):
			raise ValueError(f"The given value is not supported for StateObserverDimmer: {value}")

		self._value = value
		self._item.oh_send_command(value)


class StateObserverNumber(_StateObserverBase):
	"""Class to observe the state of a number item.

	This class is normally not used standalone. Anyway here is an example config:

	# KNX-things:
	Thing device T00_99_OpenHab_DimmerNumber "KNX OpenHAB number observer"{
        Type number             : number             "Switch"             [ ga="1/1/10" ]
    }

    # Items:
	    Number    I01_01_Number    "Switch [%s]" {channel="knx:device:bridge:T00_99_OpenHab_DimmerNumber:number"}

	# Rule init:
	habapp_rules.actors.state_observer.StateObserverNumber("I01_01_Number", callback_value_changed)
	"""

	def __init__(self, item_name: str, cb_manual: CallbackType, value_tolerance: float = 0) -> None:
		"""Init state observer for switch item.

		:param item_name: Name of switch item
		:param cb_manual: callback which should be called if manual change was detected
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		"""
		self._cb_manual = cb_manual
		_StateObserverBase.__init__(self, item_name, value_tolerance=value_tolerance)

	def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Check if light was triggered by a manual action

		:param event: event which triggered this method. This will be forwarded to the callback
		:raises ValueError: if event is not supported
		"""
		if self._value is None:
			self._value = event.value
			return

		if self._values_different_with_tolerance(event.value, self._value):
			self._value = event.value
			self._trigger_callback("_cb_manual", event)

	def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent):  # not used by StateObserverNumber
		"""Callback, which is called if a command event of one of the control items was detected.

		:param event: event, which triggered this callback
		"""

	def send_command(self, value: int | float) -> None:
		"""Send brightness command to light (this should be used by rules, to not trigger a manual action)

		:param value: Value to send to the light
		:raises ValueError: if value has wrong format
		"""
		if not isinstance(value, (int, float)):
			raise ValueError(f"The given value is not supported for StateObserverNumber: {value}")
		self._value = value
		self._item.oh_send_command(value)


class StateObserverSlat(StateObserverNumber):
	"""This is only used for the slat value of shading!"""

	def __init__(self, item_name: str, cb_manual: CallbackType, value_tolerance: float = 0) -> None:
		"""Init state observer for switch item.

		:param item_name: Name of switch item
		:param cb_manual: callback which should be called if manual change was detected
		:param value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
		"""
		self.__timer_manual: threading.Timer | None = None
		StateObserverNumber.__init__(self, item_name, cb_manual, value_tolerance)

	def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Check if light was triggered by a manual action

		:param event: event which triggered this method. This will be forwarded to the callback
		:raises ValueError: if event is not supported
		"""
		self._stop_timer_manual()
		if event.value in {0, 100}:
			self.__timer_manual = threading.Timer(3, self.__cb_check_manual_delayed, [event])
			self.__timer_manual.start()
		else:
			StateObserverNumber._check_manual(self, event)

	def __cb_check_manual_delayed(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Trigger delayed manual check

		:param event: event which should be checked
		"""
		StateObserverNumber._check_manual(self, event)

	def _stop_timer_manual(self) -> None:
		"""Stop timer if running."""
		if self.__timer_manual:
			self.__timer_manual.cancel()
			self.__timer_manual = None

	def on_rule_removed(self) -> None:
		"""Stop timer if rule is removed."""
		self._stop_timer_manual()
